/*
** Module   : ControlSetEntity.java
** Abstract : server side grouped control widget parent class
**
** Copyright (c) 2005-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 NVS 20051026   @23487 Created initial version.
** 002 SIY 20060209   @24430 Major refactoring.
** 003 GES 20060803   @28371 Some cleanup and refactoring.
** 004 GES 20061103   @30943 Fixed abend due to the use of Arrays.asList()
**                           which returns an unmodifiable list.
** 005 GES 20061108   @31026 Code and doc cleanup and shift away from the character.getValue().
** 006 GES 20061128   @31452 Fixed setListItemPairs() to add in the order specified rather than
**                           reversing the order. This only affects non-sorted lists as the
**                           sorted lists would be forced into the right order later.
** 007 SIY 20061213   @31673 Maintain dynamic pairs attribute.
** 008 SIY 20101122          Inferred generics.
** 009 SIY 20101124          Fixed UnsupportedOperationException.
** 010 SVL 20101214          Added addFirst(character, character).
** 011 ECF 20110426          Repeated fix #004 for 2 more locations.
** 012 SIY 20111101          Updated imports.
** 013 CA  20130222          Added NPE protection for cases when a widget is used before is 
**                           attached to a frame (after a CREATE WIDGET statement is used).
** 014 OM  20130527          Added int64 support fopr entry().
*/

package com.goldencode.p2j.ui;

import java.util.*;
import java.util.List;
import com.goldencode.p2j.util.*;
import com.goldencode.p2j.ui.client.format.*;

/**
 * Server side parent class for widgets that group other controls.
 */
public class ControlSetEntity
extends ControlEntity
{
   /** Reference to our configuration which is pre-cast to the proper type. */
   private ControlSetConfig csc = null;
   
   /**
    * Construct an instance using the given configuration.
    * 
    * @param   csc
    *          Source configuration.
    */
   public ControlSetEntity(ControlSetConfig csc)
   {
      super(csc);
      this.csc = csc;
   }
   
   /**
    * Implements the ENTRY() widget method.  This method returns items from
    * the set by 1-based index.
    *
    * @param    index
    *           The index of an item.
    *
    * @return   The list item indexed or <code>unknown value</code>.
    */
   @Override
   public character entry(int64 index)
   {
      if (index.isUnknown())
         return new character();
      
      int i = index.intValue();
      ControlSetItem[] items = csc.getItems();
      
      if (i < 1 || i > items.length)
         return new character();

      return items[i - 1].getLabel();
   }

   /**
    * Implements the LOOKUP() widget method.  This method returns the 1-based
    * item's index in the set.
    *
    * @param    text
    *           The text of an item.
    *
    * @return   The item's index or <code>unknown value</code>.
    */
   @Override
   public integer lookup(character text)
   {
      if (text.isUnknown())
         return new integer();

      ControlSetItem[] items = csc.getItems();
      
      for (int i = 0; i < items.length; i ++)
      {
         if (items[i].getLabel().compareTo(text) == 0)
             return new integer(i + 1);
      }

      // if list-string is not in the list, LOOKUP returns 0
      return new integer(0);
   }
   
   /**
    * Implements the LOOKUP() widget method.  This method returns the 1-based
    * item's index in the set.
    *
    * @param    text
    *           The text of an item.
    *
    * @return   The item's index or <code>unknown value</code>.
    */
   @Override
   public integer lookup(String text)
   {
      return lookup(new character(text));
   }
   
   /**
    * Implements the ADD-FIRST() widget method, which adds a list of labels
    * to the beginning of the list.
    *
    * @param    itemList
    *           A list of delimited items using the configured delimiter for
    *           this widget (a comma by default) to split the items.
    *
    * @return   <code>true</code> if the method succeeds.
    */
   @Override
   public logical addFirst(character itemList)
   {
      if (itemList.isUnknown())
         return new logical(false);

      return addFirst(itemList.toStringMessage());
   }
   
   /**
    * Implements the ADD-FIRST() widget method, which adds a list of labels
    * to the beginning of the list.
    *
    * @param    itemList
    *           A list of delimited items using the configured delimiter for
    *           this widget (a comma by default) to split the items.
    *
    * @return   <code>true</code> if the method succeeds.
    */
   @Override
   public logical addFirst(String itemList)
   {
      char[] delim = new char[1];
      delim[0] = csc.getDelimiter();
      String d = new String(delim);
      
      String[] elem = itemList.split(d);
      ControlSetItem[] items = csc.getItems();

      if (items == null)
         items = new ControlSetItem[0];

      List<ControlSetItem> list = asList(items);
      
      for (int i = elem.length - 1; i >= 0; i--)
      {
         ControlSetItem it = new ControlSetItem(new character(elem[i]),  
                                                new character(elem[i]));
         list.add(0, it); 
      }
      
      items = list.toArray(items);
      
      csc.setItems(items);
      
      pushScreenDefinition();
      return new logical(true);
   }

   /**
    * Implements the ADD-FIRST() widget method, which adds a single label
    * and value pair to the beginning of the list.
    *
    * @param    label
    *           The label of the item to add.
    * @param    value
    *           The value of the item to add.
    *
    * @return   <code>true</code> if the method succeeds.
    */
   public logical addFirst(character label, BaseDataType value)
   {
      ControlSetItem[] items = csc.getItems();
      
      if (items == null)
         items = new ControlSetItem[0];
      
      List<ControlSetItem> list = asList(items);
      list.add(0, new ControlSetItem(label, value));
      items = list.toArray(items);
      
      csc.setItems(items);
      pushScreenDefinition();
      return new logical(true);
   }
   
   /**
    * Implements the ADD-FIRST() widget method, which adds a single label
    * and value pair to the beginning of the list.
    *
    * @param    label
    *           The label of the item to add.
    * @param    value
    *           The value of the item to add.
    *
    * @return   <code>true</code> if the method succeeds.
    */
   public logical addFirst(character label, character value)
   {
      return addFirst(label, (BaseDataType) value);
   }
   
   /**
    * Store array of items into config.
    * 
    * @param    items
    *           Array of new items.
    */
   public void setItems(ControlSetItem[] items)
   {
      csc.setItems(items);
      pushScreenDefinition();
   }
 
   /**
    * Set SORT attribute.
    * 
    * @param    sort
    *           New value for the attribute.
    */
   public void setSort(logical sort)
   {
      setSort(sort.booleanValue());
   }
   
   /**
    * Set SORT attribute.
    * 
    * @param    sort
    *           New value for the attribute.
    */
   public void setSort(boolean sort)
   {
      csc.setSort(sort);
      pushScreenDefinition();
   }
   
   /**
    * Sets the LIST-ITEMS writeable attribute.
    * 
    * @param    list
    *           A list of delimited items using the configured delimiter for
    *           this widget (a comma by default) to split the items.  These
    *           items will be set as the full list of items in this widget.
    */
   @Override
   public void setListItems(character list)
   {
      if (list.isUnknown())
         return;
      
      csc.setItems(new ControlSetItem[0]);
      
      addFirst(list);
   }
   
   /**
    * Common implementation list item pair attributes (LIST-ITEM-PAIRS and
    * RADIO-BUTTONS) attribute. The parameter contains string representation
    * of all labels and values separated by a delimiter (comma is the default
    * delimiter).
    * <p>
    * Example: <code>Label1,value1,Label2,value2,...,LabelN,valueN</code>
    * 
    * @param    list
    *           A list of delimited items using the configured delimiter for
    *           this widget (a comma by default) to split the items.  These
    *           items will be set as the full list of items in this widget.
    */
   @Override
   public void setListItemPairs(character list)
   {
      if (list.isUnknown())
         return;
      
      String itemList = list.toStringMessage();
      char[] delim = new char[1];
      delim[0] = csc.getDelimiter();
      String d = new String(delim);
      String type = csc.getDataType();
      String[] elem = itemList.split(d);
      ControlSetItem[] items = csc.getItems();

      if (items == null)
         items = new ControlSetItem[0];

      List<ControlSetItem> newlist = asList(items);
      
      for (int i = 0; i < elem.length; i += 2)
      {
         BaseDataType value = null;
         
         if ((i + 1) < elem.length)
            value = DisplayFormat.instanceOfType(type, elem[i + 1]);
         else
            value = DisplayFormat.instanceOfType(type);
         
         newlist.add(new ControlSetItem(new character(elem[i]), value));
      }
      
      items = newlist.toArray(items);
      
      csc.setItems(items);
      csc.setPairs(true);
      pushScreenDefinition();
   }
   
   /**
    * Gets the LIST-ITEMS writeable attribute.
    * 
    * @return   The current value of LIST-ITEMS attribute.
    */
   @Override
   public character getListItems()
   {
      ControlSetItem[] items = csc.getItems();

      if (items == null)
         items = new ControlSetItem[0];
      
      StringBuffer sb = new StringBuffer();
      
      for (int i = 0; i < items.length; i++)
      {
         sb.append(items[i].getLabel().toStringMessage());
         
         if (i < items.length - 1)
            sb.append(csc.getDelimiter());
      }
      
      return new character(sb.toString());
   }

   /**
    * Gets the LIST-ITEM-PAIRS writeable attribute.
    *
    * @return   The current value of LIST-ITEMS attribute.
    */
   @Override
   public character getListItemPairs()
   {
      ControlSetItem[] items = csc.getItems();

      if (items == null)
         items = new ControlSetItem[0];
      
      StringBuffer sb = new StringBuffer();
      char delimiter = csc.getDelimiter(); 
      
      for (int i = 0; i < items.length; i++)
      {
         sb.append(items[i].getLabel().toStringMessage());
         sb.append(delimiter);
         sb.append(items[i].getValue().toStringMessage());
         if (i < items.length - 1)
            sb.append(delimiter);
      }
      
      return new character(sb.toString());
   }
   
   /**
    * Tell widget that data format should be honored.
    * 
    * @param    b
    *           If <code>true</code>, the format string will be honored.
    */
   public void setHonorFormat(boolean b)
   {
      csc.setHonorFormat(b);
      pushScreenDefinition();
   }
   
   /**
    * Convert the given array to a list that is modifiable.  Note that the
    * <code>Arrays.asList()</code> method does NOT return a modifiable list.
    *
    * @param    items
    *           The array to be converted.
    *
    * @return   The modifiable list in the same order as the original array.
    */
   private List<ControlSetItem> asList(ControlSetItem[] items)
   {
      ArrayList<ControlSetItem> list = new ArrayList<ControlSetItem>();
      
      for (ControlSetItem item : items)
      {
         list.add(item);
      }
      
      return list;
   }   
}
