/*
** Module   : DateFormat.java
** Abstract : date formatting and editing support
**
** Copyright (c) 2005-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- --------------------------Description---------------------------
** 001 SVG 20050609   @21446 Initial version.
** 002 SVG 20050620   @21523 DisplayFormat related classes are redesigned.
**                           Now cursor's position and insert mode are
**                           maintained inside Presentation. This allows
**                           to not pass them by reference into
**                           Presentation edit methods. All
**                           com.goldencode.p2j.ui.chui.dataref package
**                           is obsoleted and removed.
** 003 SIY 20050831   @22530 Spelling and formatting cleanups. 
** 004 SIY 20050907   @22630 Fixed use of @inheritDoc tag in javadoc.
** 005 GES 20060113   @23899 Match modified date() constructor syntax.
** 006 GES 20060503   @25945 Editing fixes. Ring terminal bell for any
**                           unexpected data keys (non-digits and non-  
**                           separators).  Clear last char state for
**                           each new data key so that if true it is
**                           clearly a new condition rather than one that
**                           is carried over from prior key processing.
** 007 GES 20060503   @25954 Added edit mode to rendering where an empty
**                           string with the separators is returned 
**                           instead of an empty string when edit mode is 
**                           true.
** 008 GES 20060508   @26017 Fix for format checking to catch correct 
**                           exception type.
** 009 GES 20060724   @28153 Use forced assignment mode of BaseDataType
**                           to avoid checking the ErrorManager for a 
**                           pending error (that causes a trip to the
**                           server which is unnecessary).
** 010 GES 20060918   @29642 Removed unnecessary limit on years before
**                           0 and after 9999. Completely rewrote the
**                           internal data storage mechanism to ensure
**                           that negative year values are handled.
** 011 GES 20060920   @26729 Added support for isUnknown().
** 012 SVL 20080226   @37174 Year autocompletion mechanism and
**                           string-to-year conversion reworked. Testcases
**                           added.
** 013 GES 20080410   @37915 Initialize the date order using the date
**                           class which reads this from the server's
**                           directory if needed. This allows a context
**                           specific value to be shared for both the
**                           client and server. The date component order
**                           array is compatible between the date class
**                           and this one. Both classes honor the ordering
**                           definition.
** 014 SIY 20080606   @38614 Clearing date should behave as if date was
**                           initialized with unknown value.
**                           Ref: #38613
** 015 SVL 20080922   @39894 Insert mode reset removed from the
**                           activate().
** 016 SVL 20090120   @41184 Added drawSeparators flag.
** 017 SVL 20090402   @41719 Insert mode is saved on clear().
** 018 SIY 20090701   @43020 Added DateBuf.validateFormat() implementation.
** 019 SIY 20090721   @43290 Moved activate() into base class.
** 020 SIY 20100619          Switching to own implementation of character UI.
** 021 SIY 20100804          Changes in imports, some refactorings.
** 022 SIY 20111101          Updated imports.
** 023 OM  20130329          Use date.today() instead of constructor. Cosmetical changes.
*/

package com.goldencode.p2j.ui.client.format;

import static com.goldencode.p2j.util.date.*;

import java.util.*;

import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.util.*;
import com.goldencode.p2j.util.ErrorManager.ErrorEntry;

/**
 * Display format for <code>date</code> variables.
 */
public class DateFormat
extends DisplayFormat
{
   // TODO: already existing instances won't pick up the latest change
   //       when SESSION:DATE-FORMAT is called, but in the 4GL they will
   /** Order of date components for rendering. */
   protected byte[] dateCompOrder = date.getDateComponentOrder();

   /** Date components separator. */
   protected char fmtSeparator;

   /**
    * Determines whether the separators will be drawn for the unknown value
    * (applies to editing mode only). 
    */
   protected boolean drawSeparators = true;

   /** <code>true</code> for a 4-digit year. */
   protected boolean fmtFullYear;

   /**
    * How date components are positioned in rendered string. This array is
    * directly indexed by the component's code constants (DAY etc.).
    */
   protected DateComponent[] fmtDateComps;

   /** Default width which should be reserved for a screen value. */
   private int fmtScreenWidth;

   /**
    * Constructs a date format from the given Progress format string.
    * 
    * @param   pformat
    *          Progress format string.
    * @throws  DisplayFormatParsingException
    *          When passed format is not a valid date format.
    */
   public DateFormat(String pformat) 
   throws DisplayFormatParsingException
   {
      super(pformat);

      // TODO: this is a bad approach because it does not support a wide
      //       range of perfectly valid date formats, FIX THIS by making it
      //       generic (e.g. "99/99/-99999" or "99/99/999" or "99-99/99")
      
      if (pformat.equals("99/99/99"))
      {
         fmtSeparator = '/';
         fmtFullYear = false;
      }
      else if (pformat.equals("99/99/9999"))
      {
         fmtSeparator = '/';
         fmtFullYear = (dateCompOrder[2] == YEAR);
      }
      else if (pformat.equals("99/9999/99"))
      {
         fmtSeparator = '/';
         fmtFullYear = (dateCompOrder[1] == YEAR);
      }
      else if (pformat.equals("9999/99/99"))
      {
         fmtSeparator = '/';
         fmtFullYear = (dateCompOrder[0] == YEAR);
      }
      else if (pformat.equals("99-99-99"))
      {
         fmtSeparator = '-';
         fmtFullYear = false;
      }
      else if (pformat.equals("99-99-9999"))
      {
         fmtSeparator = '-';
         fmtFullYear = (dateCompOrder[2] == YEAR);
      }
      else if (pformat.equals("99-9999-99"))
      {
         fmtSeparator = '-';
         fmtFullYear = (dateCompOrder[1] == YEAR);
      }
      else if (pformat.equals("9999-99-99"))
      {
         fmtSeparator = '-';
         fmtFullYear = (dateCompOrder[0] == YEAR);
      }
      else if (pformat.equals("99.99.99"))
      {
         fmtSeparator = '.';
         fmtFullYear = false;
      }
      else if (pformat.equals("99.99.9999"))
      {
         fmtSeparator = '.';
         fmtFullYear = (dateCompOrder[2] == YEAR);
      }
      else if (pformat.equals("99.9999.99"))
      {
         fmtSeparator = '.';
         fmtFullYear = (dateCompOrder[1] == YEAR);
      }
      else if (pformat.equals("9999.99.99"))
      {
         fmtSeparator = '.';
         fmtFullYear = (dateCompOrder[0] == YEAR);
      }
      else if (pformat.equals("999999"))
      {
         fmtSeparator = 0;
         fmtFullYear = false;
      }
      else if (pformat.equals("99999999"))
      {
         fmtSeparator = 0;
         fmtFullYear = true;
      }
      else
      {
         throw new DisplayFormatParsingException("Invalid date format: " + pformat);
      }

      fmtDateComps = new DateComponent[3];
      int offs = 0;
      for (int i = 0; i < 3; i++)
      {
         int compIndex = dateCompOrder[i];
         int length = (compIndex == YEAR && fmtFullYear) ? 4 : 2;
         fmtDateComps[compIndex] = new DateComponent(offs, length);
         offs += length;
         
         if (fmtSeparator != 0)
         {
            offs++;
         }
      }

      fmtScreenWidth = 6;
      if (fmtSeparator != 0)
      {
         fmtScreenWidth += 2;
      }
      if (fmtFullYear)
      {
         fmtScreenWidth += 2;
      }
   }

   /**
    * Returns screen width of the format.  The screen width is the number of
    * characters that will be rendered.  This is used as the default size for
    * an editor.
    * 
    * @return   The screen width of the format. 
    */
   @Override
   public int getScreenWidth()
   {
      return fmtScreenWidth;
   }

   /**
    * Returns a new presentation instance with a default initialized insert
    * mode and cursor position.
    * 
    * @param   pres
    *          A Presentation instance which was constructed by this format.
    *          May be <code>null</code>. This presentation instance is not 
    *          changed in this method.
    * 
    * @return  New presentation instance.
    */
   @Override
   public Presentation clear(Presentation pres)
   {
      drawSeparators = false;

      DateBuf buf = new DateBuf();
      
      buf.compBufs = null;

      if (pres != null)
      {
         buf.presInsertMode = pres.presInsertMode;
      }
      
      return buf;
   }

   /**
    * Initializes a presentation for variable's value.
    * 
    * @param   var
    *          Variable which should be compatible with format.
    * 
    * @return  Data presentation
    * 
    * @throws  DisplayFormatCheckException
    *          Variable's value is incompatible with format.
    */
   @Override
   public Presentation fromVar(BaseDataType var)
   throws DisplayFormatCheckException
   {
      return new DateBuf((date)var);
   }

   /**
    * Forces separators for the unknown date value to be drawn or not in the
    * subsequent value rendering (applies to editing mode only).
    *
    * @param drawSeparators
    *        If <code>true</code> then separators will be displayed for the
    *        unknown date value.
    */
   public void setDrawSeparators(boolean drawSeparators)
   {
      this.drawSeparators = drawSeparators;
   }

   /**
    * The simple structure which records what places are occupied in a screen
    * value by the date component.
    */
   private static class DateComponent
   {
      /** Offset of this component in rendered string. */
      private int compOffset;

      /** Number of digits to render. */
      private int compLength;

      /**
       * Constructor.
       * 
       * @param   offset
       *          Offset of this component in rendered string.
       * @param   length
       *          Number of digits to render.
       */
      public DateComponent(int offset, int length)
      {
         compOffset = offset;
         compLength = length;
      }

      /**
       * Returns offset of this component in rendered string.
       * 
       * @return  Offset of this component in rendered string.
       */
      public int getOffset()
      {
         return compOffset;
      }

      /**
       * Returns number of digits to render.
       * 
       * @return  Number of digits to render.
       */
      public int getLength()
      {
         return compLength;
      }
   }

   /**
    * Internal form of date data.
    */
   private class DateBuf
   extends Presentation
   {
      /** Special value for an empty digit place. */
      private static final char SPACE = ' ';

      /**
       * Buffers for date component digits. May be <code>null</code> which
       * means UNKNOWN value. Array element must contain a digit or
       * <code>SPACE</code> which will be rendered as space.
       */
      private char[][] compBufs = null;

      /**
       * Constructs default presentation used for clear operation.
       */
      public DateBuf()
      {
         presCursorPos = 0;
         presInsertMode = false;
         initClear();
      }

      /**
       * Constructs presentation initialized with given <code>date</code>.
       * 
       * @param   varDate
       *          Initial value.
       * 
       * @throws  DisplayFormatCheckException
       *          This value can't be displayed with the current format.
       */
      public DateBuf(date varDate)
      throws DisplayFormatCheckException
      {
         if (varDate.isUnknown())
         {
            return;
         }

         int day   = varDate.getDay();
         int month = varDate.getMonth();
         int year  = varDate.getYear();
         
         if (!fmtFullYear)
         {
            if (!varDate.inY2KWindow())
            {
               ErrorManager.displayError(74,
                     "Value " + varDate.toStringMessage() + 
                     " cannot be displayed using " + getFormatDef());
               
               //throw new DisplayFormatCheckException(
               //   "Value " +
               //   varDate.toStringMessage() +
               //   " cannot be displayed using " +
               //   getFormatDef() +
               //   ". Year must be in current century.");
            }

            year %= 100;
         }

         compBufs = initBuffers();
         assignComponent(compBufs[DAY], day);
         assignComponent(compBufs[MONTH], month);
         assignComponent(compBufs[YEAR], year);
      }

      /**
       * Checks format rules. Method finishEdit() should be called before it.
       * 
       * @throws  DisplayFormatCheckException
       *          When data is incompatible with format
       */
      @Override
      public void checkFormat()
      throws DisplayFormatCheckException
      {
         // just call this to check for exceptions, discard the return value
         toDate();
      }

      /**
       * Validate representation according to format rules and return error
       * message if there is an error.
       * 
       * @return   <code>null</code> if no error or error message if there is an
       *           error.
       */
      @Override
      public ErrorEntry validateFormat()
      {
         int day = -1;
         int month = -1;
         
         // assume unedited date is valid
         if (compBufs == null)
            return null;
         
         try
         {
            day = extractDayOrMonthComponent(compBufs[DAY]);
            month = extractDayOrMonthComponent(compBufs[MONTH]);
         }
         catch (NumberFormatException e)
         {
            // error will be handled later
         }

         ErrorEntry res = null;
         
         res = verifyYearComponent(compBufs);
         
         if (res != null)
            return res;

         try
         {
            return date.validate(month, 
                                 day, 
                                 extractYearComponent(compBufs), 
                                 !fmtFullYear);
         }
         catch (ErrorConditionException ex)
         {
            // should not happen because we already verified data
         }
         
         return null;
      }
      
      /**
       * Returns if the presentation currently represents the 
       * <code>unknown value</code>.  If any editing has occurred, then this
       * will return <code>false</code> even if the editing is not complete.
       *
       * @return   <code>true</code> if this is <code>unknown</code>.
       */
      @Override
      public boolean isUnknown()
      {
         return (compBufs == null);
      }

      /**
       * Erases one symbol before the cursor position.
       * 
       * @return   <code>true</code> if input is accepted (nevertheless it
       *           may be silently ignored in reality).
       */
      @Override
      public boolean deleteBack()
      {
         if (presCursorPos > 0)
         {
            presCursorPos--;
            return true;
         }
         
         return false;
      }

      /**
       * Erases one symbol in the cursor position.
       * 
       * @return   <code>true</code> if input is accepted (nevertheless it
       *           may be silently ignored in reality).
       */
      @Override
      public boolean deleteCurrent()
      {
         return false;
      }

      /**
       * Inputs character. The result depends from the current insert mode.
       * 
       * @param   ch
       *          Input character
       * 
       * @return   <code>true</code> if input is accepted (nevertheless it
       *           may be silently ignored in reality).
       */
      @Override
      public boolean input(char ch)
      {
         if (presCursorPos >= getScreenWidth())
         { 
            // ignore any input outside widget even '?'
            return false;
         }

         if (ch == '?')
         {
            compBufs = null;
            presCursorPos = 0;
            drawSeparators = false;
            return true;
         }

         if (compBufs == null)
         { 
            // restart method from zero content
            initClear();
            boolean res = input(ch);
            if (!res)
            {
               compBufs = null;
            }
            return res;
         }

         int cursorPos = presCursorPos;
         
         // clear flag for last char
         setLastCharEntered(false);
         
         // iterate through date components
         for (int i = 0; i < dateCompOrder.length; i++)
         {
            byte compIndex = dateCompOrder[i];
            DateComponent comp = fmtDateComps[compIndex];
            int offset = comp.getOffset();

            // bypass components which we are not in
            if (cursorPos >= offset + comp.getLength())
            {
               continue;
            }

            byte digit = (byte) Character.digit(ch, 10);
            int pos = cursorPos - offset;
            if (digit >= 0)
            {
               if (pos < 0)
               {
                  // skip separator
                  pos = 0;
                  cursorPos = offset;
               }

               compBufs[compIndex][pos] = ch;

               // is this the last digit position of the last component?
               if (i == dateCompOrder.length - 1 &&
                   pos == compBufs[compIndex].length - 1)
               {
                  setLastCharEntered(true);
               }

               break;
            }

            if (ch == '-'          &&
                compIndex == YEAR  &&
                pos >= 0           &&
                fmtSeparator != 0)
            {
               compBufs[compIndex][pos] = ch;

               // is this the last digit position of the last component?
               if (i == dateCompOrder.length - 1 &&
                   pos == compBufs[compIndex].length - 1)
               {
                  setLastCharEntered(true);
               }

               break;
            }

            if (ch == fmtSeparator)
            {
               if (pos < 0)
               {
                  // Separator char entered over existing separator.
                  // Only moving cursor
                  break;
               }

               // completing current component and moving to next if possible
               completeComponent(compBufs, compBufs, compIndex);
               if (i < dateCompOrder.length - 1)
               {
                  // next component available. Separator is always present in
                  // this case so we can calculate offset with only current
                  // component
                  presCursorPos = offset + comp.getLength() + 1;
               }
               else
               { 
                  // no move
                  presCursorPos = cursorPos;
               }
               return true;
            }

            // invalid input rings the terminal bell
            OutputManager.instance().beep();
            
            return false;
         }

         // move the cursor right even if we go past the end by 1
         presCursorPos = cursorPos + 1;

         return true;
      }

      /**
       * Moves cursor one character to the left.
       * 
       * @return   <code>true</code> if input is accepted (nevertheless it
       *           may be silently ignored in reality).
       */
      @Override
      public boolean moveCursorLeft()
      {
         if (presCursorPos > 0)
         {
            presCursorPos--;
         }
         return true;
      }

      /**
       * Moves cursor one character to the right. Also updates insert mode.
       * 
       * @return   <code>true</code> if input is accepted (nevertheless it
       *           may be silently ignored in reality).
       */
      @Override
      public boolean moveCursorRight()
      {
         if (presCursorPos < getScreenWidth() - 1)
         {
            presCursorPos++;
         }
         return true;
      }

      /**
       * Moves cursor to the beginning of the field.
       * 
       * @return   <code>true</code> if input is accepted (nevertheless it
       *           may be silently ignored in reality).
       */
      @Override
      public boolean moveCursorBegin()
      {
         presCursorPos = 0;
         return true;
      }

      /**
       * Moves cursor to the end of the field.
       * 
       * @return   <code>true</code> if input is accepted (nevertheless it
       *           may be silently ignored in reality).
       */
      @Override
      public boolean moveCursorEnd()
      {
         presCursorPos = getScreenWidth() - 1;
         return true;
      }

      /**
       * Indicates completion of an edit. The presentation instance may fill 
       * some unfilled places.
       * 
       * @return   <code>true</code> if rendering is necessary.
       */
      @Override
      public boolean finishEdit()
      {
         if (compBufs == null)
         {
            return false;
         }

         char[][] newBufs = initBuffers();

         boolean changed = false;
         
         for (int i = 0; i < 3; i++)
         {
            completeComponent(newBufs, compBufs, i);
            if (!Arrays.equals(newBufs[i], compBufs[i]))
            {
               changed = true;
            }
         }

         if (changed)
         {
            compBufs = newBufs;
         }

         return changed;
      }

      /**
       * Formats the data as a string that is ready for display.
       * 
       * @param    edit
       *           <code>true</code> if the screen value is being generated
       *           for interactive editing purposes.
       * 
       * @return   The formatted string.
       */
      @Override
      public String toScreenValue(boolean edit)
      {
         boolean tempBuf = (edit && (compBufs == null) && drawSeparators);

         if (tempBuf)
         {
            // create temporary buffers
            initClear();
         }
         
         String result = toScreenValue();
         
         if (tempBuf)
         {
            // remove temporary buffers
            compBufs = null;
         }
         
         return result;
      }

      /**
       * Formats the data as a string that is ready for display.
       * 
       * @return   The formatted string.
       */
      @Override
      public String toScreenValue()
      {
         if (compBufs == null)
         {
            return "";
         }

         StringBuffer res = new StringBuffer(10);
         for (int i = 0; i < dateCompOrder.length; i++)
         {
            if (i > 0 && fmtSeparator != 0)
            {
               res.append(fmtSeparator);
            }

            res.append(compBufs[dateCompOrder[i]]);
         }

         return res.toString();
      }

      /**
       * Instantiates a variable that is equivalent to the presentation's
       * internal representation.
       * 
       * @param    var
       *           A variable which must be compatible with the format.
       */
      @Override
      public void toVar(BaseDataType var)
      {
         try
         {
            ((date) var).assign(toDate(), true);
         }
         catch (DisplayFormatCheckException ex)
         {
            // buffers should be already validated
         }
      }

      /**
       * Allocates date components buffers.
       * 
       * @return  Components buffers array.
       */
      private char[][] initBuffers()
      {
         char[][] bufs = new char[3][];
         bufs[DAY]   = new char[2];
         bufs[MONTH] = new char[2];
         bufs[YEAR]  = new char[(fmtFullYear) ? 4 : 2];
         return bufs;
      }

      /**
       * Resets content to space value.
       */
      private void initClear()
      {
         compBufs = initBuffers();
         for (int i = 0; i < 3; i++)
         {
            char[] compBuf = compBufs[i];
            clearComponent(compBuf);
         }
      }

      /**
       * Fills the date component with <code>SPACE</code>.
       * 
       * @param   comp
       *          The component.
       */
      private void clearComponent(char[] comp)
      {
         for (int i = 0; i < comp.length; ++i)
         {
            comp[i] = SPACE;
         }
      }

      /**
       * Fills the date component with the given value.
       * 
       * @param   comp
       *          The component.
       * @param   value
       *          The new value.
       */
      private void assignComponent(char[] comp, int value)
      {
         String txt = Integer.toString(value);
         
         int idx = txt.length() - 1;
         
         for (int i = comp.length - 1; i >= 0; i--)
         {
            if (idx >= 0)
            {
               comp[i] = txt.charAt(idx);
               idx--;
            }
            else
            {
               // left pad
               comp[i] = '0';
            }
         }
      }

      /**
       * Calculates numeric value of the date component.
       * 
       * @param   comp
       *          The date component (day or month). May partially consist from spaces.
       *
       * @return  Not negative integer value of component. (-1) if it is empty.
       */
      private int extractDayOrMonthComponent(char[] comp)
      {
         int begin = 0;

         while (begin < comp.length)
         {
            if (comp[begin] != SPACE)
            {
               break;
            }
            begin++;
         }

         if (begin == comp.length)
         {
            return -1;
         }

         int end = begin + 1;
         while (end < comp.length)
         {
            if (comp[end] == SPACE)
            {
               break;
            }
            end++;
         }

         String txt = new String(comp);
         
         return Integer.parseInt(txt.substring(begin, end));
      }

      /**
       * Calculates numeric value of the year. Calculation can be described by
       * the following algorithm:
       * <p>
       * If format separator isn't <code>null</code> then:
       * <p>
       * <ol>
       *    <li> Input sequence is autocompleted (to a temporary buffer).
       *    <li> If the extreme right position holds '-' then it is discarded
       *         (e.g. -10- -> -10)
       *    <li> Remaining characters are parsed as a usual integer.
       * </ol>
       * <p>
       * If format separator is <code>null</code> then:
       * <p>
       * <ol>
       *    <li> Only the extreme left (before the space if it presents)
       *         digits are taken into account. If the first character is a
       *         space then the whole field content is ignored (e.g. 12_3 ->
       *         12; _123 -> <code>null</code>)
       *    <li> If input consists of three digits then the last one is
       *         ignored (e.g. 123_ -> 12).
       *    <li> Remaining characters are parsed as a usual integer. If it
       *         cannot be parsed then the current year is used.
       * </ol>
       *
       * @param   comp
       *          The input date buffer.
       *
       * @return  The numeric value of the year.
       *
       * @throws  ErrorConditionException
       *          If the input sequence does not represent a valid year.
       */
      private int extractYearComponent(char[][] comp)
      throws ErrorConditionException
      {
         char[][] tempBuf = new char[3][];
         tempBuf[YEAR] = Arrays.copyOf(comp[YEAR], comp[YEAR].length);
         for (char c : comp[YEAR])
         {
            if (c == SPACE)
            {
               // process autocompletion for a temporary buffer if it is
               // required
               completeComponent(tempBuf, comp, YEAR);
               break;
            }
         }

         char[] yearBuffer = tempBuf[YEAR];

         int year = 0;
         if (fmtSeparator != 0)
         {
            int end = yearBuffer.length - 1;

            // discard extreme right '-' if present
            if (yearBuffer[end] == '-')
            {
               end--;
            }

            String txt = new String(yearBuffer);
            try
            {
               year = Integer.parseInt(txt.substring(0, end + 1));
            }
            catch(NumberFormatException ex)
            {
               if(fmtFullYear)
               {
                  ErrorManager.recordOrThrowError(85, "Invalid date input");
                  return (int)date.INVALID_DATE;
               }
            }
         }
         else // fmtSeparator == 0
         {
            // Only extreme left (before the space if it presents) digits
            // are taken into account. If the first character is a space
            // then the whole field content is ignored.
            int end = yearBuffer.length - 1;
            for (int i = 0; i < yearBuffer.length; i++)
            {
               if (yearBuffer[i] == SPACE)
               {
                  end = i - 1;
                  break;
               }
            }

            String txt = new String(yearBuffer);
            txt = txt.substring(0, end + 1);

            // if input consists of three digits then the last one is ignored
            if (txt.length() == 3)
            {
               txt = txt.substring(0, 2);
            }

            try
            {
               year = Integer.parseInt(txt);
            }
            catch(NumberFormatException ex)
            {
               year = date.today().getYear();
            }
         }

         if (year == 0 && fmtFullYear)
         {
            ErrorManager.recordOrThrowError(79, "Year is out of range of 0");
            return (int) date.INVALID_DATE;
         }

         return year;
      }

      /**
       * Validate year without throwing exceptions.
       * 
       * @param    comp
       *           Component to verify.
       * 
       * @return   <code>null</code> if no error found, {@link ErrorEntry}
       *           instance otherwise.
       */
      private ErrorEntry verifyYearComponent(char[][] comp)
      {
         char[][] tempBuf = new char[3][];
         tempBuf[YEAR] = Arrays.copyOf(comp[YEAR], comp[YEAR].length);
         for (char c : comp[YEAR])
         {
            if (c == SPACE)
            {
               // process autocompletion for a temporary buffer if it is
               // required
               completeComponent(tempBuf, comp, YEAR);
               break;
            }
         }

         char[] yearBuffer = tempBuf[YEAR];

         int year = 0;
         if (fmtSeparator != 0)
         {
            int end = yearBuffer.length - 1;

            // discard extreme right '-' if present
            if (yearBuffer[end] == '-')
            {
               end--;
            }

            String txt = new String(yearBuffer);
            try
            {
               year = Integer.parseInt(txt.substring(0, end + 1));
            }
            catch (NumberFormatException ex)
            {
               if (fmtFullYear)
               {
                  return new ErrorEntry(85, "Invalid date input");
               }
            }
         }
         else // fmtSeparator == 0
         {
            // Only extreme left (before the space if it presents) digits
            // are taken into account. If the first character is a space
            // then the whole field content is ignored.
            int end = yearBuffer.length - 1;
            for (int i = 0; i < yearBuffer.length; i++)
            {
               if (yearBuffer[i] == SPACE)
               {
                  end = i - 1;
                  break;
               }
            }

            String txt = new String(yearBuffer);
            txt = txt.substring(0, end + 1);

            // if input consists of three digits then the last one is ignored
            if (txt.length() == 3)
            {
               txt = txt.substring(0, 2);
            }

            try
            {
               year = Integer.parseInt(txt);
            }
            catch (NumberFormatException ex)
            {
               year = date.today().getYear();
            }
         }

         if (year == 0 && fmtFullYear)
         {
            return new ErrorEntry(79, "Year is out of range of 0");
         }

         return null;
      }
      
      /**
       * Completes date component. Day and month are completed with 0's.
       * <p>
       * This method uses only one old and one new buffer for component but
       * takes all buffers to simplify coding of its call.
       * <p>
       * The year completion algorithm can be described as follows:
       * <p>
       * <ol>
       *    <li> If format separator is <code>null</code> then do nothing.
       *    <li> Find the extreme right '-' and fill all empty placeholders
       *         to the left of it with the '-' (e.g. __-_ -> ---_; _1_- ->
       *         -1--).
       *    <li> The extreme leftmost meaningful symbol (not SPACE) is found.
       *         After that we search for the extreme right meaningful symbol
       *         such that no spaces are present between it and the extreme
       *         left meaningful symbol. All symbols which are not between
       *         these ones are discarded (e.g. __12 -> 12; -2_3 -> -2;
       *         _1_3 -> 1). If no meaningful symbols are present then the
       *         current year is used.
       *    <li> To the previously found sequence of meaningful symbols some
       *         leading symbol may be added depending on the sequence length
       *         and year format:
       *    <ul>
       *       <li> When length is 3 and format is full-year OR
       *            length is 1 and format is not full-year then add a
       *            leading 0.
       *       <li> When length is 2 or 1 and format is full-year then the
       *            date.windowingImpl() which takes as input the digits from
       *            the sequence (the one that was created on the third step)
       *            is used (if the sequence contains no digits then 0 is
       *            used).
       *       <li> From the return value we take first 2 or 3 characters and
       *            complete the sequence with it. (e.g. -2- -> 0-2-; 99 ->
       *            1999; -- -> 20--; 9 -> 2009)
       *     </ul>
       * </ol>
       *
       * @param   newCompBufs
       *          Buffers for new value. May be the same with
       *          <code>oldCompBufs</code>.
       * @param   oldCompBufs
       *          Buffers for old value.
       * @param   compIndex
       *          Component's index (DAY, MONTH, YEAR).
       */
      private void completeComponent(char[][] newCompBufs,
                                     char[][] oldCompBufs,
                                     int      compIndex)
      {
         if (compIndex != YEAR)
         {
            int comp = extractDayOrMonthComponent(oldCompBufs[compIndex]);

            if (comp < 0)
            {
               comp = 0;
            }

            assignComponent(newCompBufs[compIndex], comp);
         }
         else
         {
            // compIndex == YEAR
            
            if (fmtSeparator == 0)
            {
               newCompBufs[compIndex] =
                  Arrays.copyOf(oldCompBufs[compIndex],
                                oldCompBufs[compIndex].length);
               return;
            }

            char[] yearBuffer =
               Arrays.copyOf(oldCompBufs[compIndex],
                             oldCompBufs[compIndex].length);

            // find extreme right '-'
            int rightMinusPos = -1;
            for (int i = yearBuffer.length - 1; i >= 0; i--)
            {
               if (yearBuffer[i] == '-')
               {
                  rightMinusPos = i;
                  break;
               }
            }

            if (rightMinusPos != -1)
            {
               // fill all empty placeholders to the left of the
               // extreme right '-' with '-'

               for (int i = rightMinusPos - 1; i >= 0; i--)
               {
                  if (yearBuffer[i] == SPACE)
                  {
                     yearBuffer[i] = '-';
                  }
               }
            }

            // find extreme left meaningful symbol (not SPACE)
            int leftSymbolPos = -1;
            for (int i = 0; i < yearBuffer.length; i++)
            {
               if (yearBuffer[i] != SPACE)
               {
                  leftSymbolPos = i;
                  break;
               }
            }

            if (leftSymbolPos == -1)
            {
               // empty year - fill with the current one
               assignComponent(newCompBufs[compIndex], date.today().getYear());
               return;
            }

            // find extreme right meaningful symbol between which and the
            // extreme left meaningful symbol no spaces present
            int rightSymbolPos = leftSymbolPos;
            for (int i = leftSymbolPos + 1; i < yearBuffer.length; i++)
            {
               if (yearBuffer[i] != SPACE)
               {
                  rightSymbolPos = i;
               }
               else
               {
                  break;
               }
            }

            // length of the meaningful symbols sequence
            // all other symbols which are not present in this sequence
            // will not be taken into account any further
            int symbolsLength = rightSymbolPos - leftSymbolPos + 1;

            // move meaningful symbols to the right
            int writePos = yearBuffer.length - 1;
            for (int readPos = rightSymbolPos;
                 readPos >= leftSymbolPos;
                 readPos--)
            {
               yearBuffer[writePos--] = yearBuffer[readPos];
            }

            if (fmtFullYear)
            {
               // add leading year numbers
               if (symbolsLength == 3)
               {
                  yearBuffer[0] = '0';
               }
               else if (symbolsLength == 2 || symbolsLength == 1)
               {
                  // add two leading year numbers
                  StringBuffer lastYearDigits = new StringBuffer("");
                  if (symbolsLength == 2 &&
                      Character.isDigit(yearBuffer[2]))
                  {
                     lastYearDigits.append(yearBuffer[2]);
                  }
                  if (Character.isDigit(yearBuffer[3]))
                  {
                     lastYearDigits.append(yearBuffer[3]);
                  }

                  int lastYearDigitsInt = 0;
                  if (lastYearDigits.length() > 0)
                  {
                     lastYearDigitsInt =
                        Integer.parseInt(lastYearDigits.toString());
                  }

                  String fullYear =
                     Integer.toString(date.windowingImpl(lastYearDigitsInt));

                  if (fullYear.length() == 4)
                  {
                     for (int i = 0;
                          i < yearBuffer.length - symbolsLength;
                          i++)
                     {
                        yearBuffer[i] = fullYear.charAt(i);
                     }
                  }
                  else
                  {
                     // just in case
                     for (int i = 0;
                          i < yearBuffer.length - symbolsLength;
                          i++)
                     {
                        yearBuffer[i] = '0';
                     }
                  }
               }
            }
            else // not fmtFullYear
            {
               if (symbolsLength == 1)
               {
                  yearBuffer[0] = '0';
               }
            }

            newCompBufs[compIndex] = Arrays.copyOf(yearBuffer, yearBuffer.length);
         }
      }

      /**
       * Returns the current presentation's value converted to <code>date</code>.
       * 
       * @return  The current presentation's value converted to <code>date</code>.
       * 
       * @throws  DisplayFormatCheckException
       *          Value doesn't represent a correct date.
       */
      private date toDate()
      throws DisplayFormatCheckException
      {
         if (compBufs == null)
         {
            return new date(); // unknown date
         }

         // just extracting components. Spaces should be already eliminated
         int day   = extractDayOrMonthComponent(compBufs[DAY]);
         int month = extractDayOrMonthComponent(compBufs[MONTH]);

         try
         {
            int year = extractYearComponent(compBufs);
            return new date(month, day, year, !fmtFullYear);
         }
         catch (ErrorConditionException ex)
         {
            throw new DisplayFormatCheckException(ex.getMessage());
         }
      }
   }

   /**
    * Simple testcase wrapper to check the year autocompletion and its
    * string-to-number conversion.
    */
   private class TestCase
   {
      /**
       * Input sequence of characters entered by one in a year part of a date
       * field.
       */
      private String oldBuf;

      /**
       * Input sequence of characters after the autocompletion mechanism has
       * worked in it.
       */
      private String newBuf;

      /** The year that input sequence should be converted to. */
      private Integer year;


      /**
       * Construct an instance of a test case with all data needed to test
       * the year autocompletion and its string-to-number conversion.
       *
       * @param    oldBuf
       *           Input sequence of characters entered by one in a year part
       *           of a date field. The '_' symbol in this sequence means an
       *           empty placeholder.
       * @param    newBuf
       *           Input sequence of characters after the autocompletion
       *           mechanism has worked in it. The '_' symbol in this sequence
       *           means an empty placeholder.
       * @param    year
       *           The year that input sequence should be converted to.
       *           <code>null</code> means an invalid year.
       */
      public TestCase(String oldBuf, String newBuf, Integer year)
      {
         this.oldBuf = oldBuf.replaceAll("_"," ");
         this.newBuf = newBuf.replaceAll("_"," ");
         this.year = year;
      }

      /**
       * Get the input sequence of characters.
       *
       * @return   Input sequence of characters.
       */
      public String getOldBuf()
      {
         return oldBuf;
      }

      /**
       * Get the input sequence of characters after autocompletion.
       *
       * @return   Input sequence of characters after autocompletion.
       */
      public String getNewBuf()
      {
         return newBuf;
      }

      /**
       * Get the year that input sequence should be converted to.
       *
       * @return   The year that input sequence should be converted to.
       *           <code>null</code> if the sequence represents an invalid
       *           year. 
       */
      public Integer getYear()
      {
         return year;
      }
   }

   /**
    * Execute all year autocompletion and its string-to-number conversion
    * testcases.
    */
   public void test()
   {
      TestCase[] tests =
      {
         new TestCase("11_-", "11--", null),
         new TestCase("10_-", "10--", null),
         new TestCase("--_-", "----", null),
         new TestCase("--_0", "20--", null),
         new TestCase("--_1", "20--", null),
         new TestCase("1-_-", "1---", null),
         new TestCase("1-_0", "201-", 201),
         new TestCase("1-_1", "201-", 201),
         new TestCase("00_-", "00--", null),
         new TestCase("01_-", "01--", null),
         new TestCase("-0_-", "-0--", null),
         new TestCase("-0_0", "20-0", null),
         new TestCase("-0_1", "20-0", null),
         new TestCase("-1_-", "-1--", null),
         new TestCase("0-_-", "0---", null),
         new TestCase("0-_0", "200-", 200),
         new TestCase("0-_1", "200-", 200),
         new TestCase("-_--", "----", null),
         new TestCase("-_-0", "---0", null),
         new TestCase("-_-1", "---1", null),
         new TestCase("-_0-", "--0-", null),
         new TestCase("-_00", "200-", 200),
         new TestCase("-_01", "200-", 200),
         new TestCase("-_1-", "--1-", null),
         new TestCase("-_10", "200-", 200),
         new TestCase("-_11", "200-", 200),
         new TestCase("1_-_", "01--", null),
         new TestCase("1_--", "1---", null),
         new TestCase("1_-0", "1--0", null),
         new TestCase("1_-1", "1--1", null),
         new TestCase("1_0-", "1-0-", null),
         new TestCase("1_1-", "1-1-", null),
         new TestCase("0_--", "0---", null),
         new TestCase("0_-0", "0--0", null),
         new TestCase("0_-1", "0--1", null),
         new TestCase("0_0-", "0-0-", null),
         new TestCase("0_1-", "0-1-", null),
         new TestCase("1__-", "1---", null),
         new TestCase("___-", "----", null),
         new TestCase("__-_", "0---", null),
         new TestCase("_-__", "20--", null),
         new TestCase("_-_-", "----", null),
         new TestCase("_-_0", "20--", null),
         new TestCase("_-_1", "20--", null),
         new TestCase("_0_-", "-0--", null),
         new TestCase("_1_-", "-1--", null),
         new TestCase("_1-_", "0-1-", null),
         new TestCase("-___", "200-", 200),
         new TestCase("-__-", "----", null),
         new TestCase("-__0", "200-", 200),
         new TestCase("-__1", "200-", 200),
         new TestCase("-_-_", "0---", null),
         new TestCase("-_0_", "200-", 200),
         new TestCase("-_1_", "200-", 200),
         new TestCase("-1_0", "20-1", null),
         new TestCase("-1_1", "20-1", null),
         new TestCase("-1-_", "0-1-", null),
         new TestCase("0__-", "0---", null),
         new TestCase("0_-_", "00--", null),
         new TestCase("____", "2008", 2008),
         new TestCase("0-0-", "0-0-", null),
         new TestCase("0-1-", "0-1-", null),
         new TestCase("01--", "01--", null),
         new TestCase("0---", "0---", null),
         new TestCase("-0--", "-0--", null),
         new TestCase("-0-0", "-0-0", null),
         new TestCase("-0-1", "-0-1", null),
         new TestCase("--1-", "--1-", null),
         new TestCase("--10", "--10", null),
         new TestCase("--11", "--11", null),
         new TestCase("--0-", "--0-", null),
         new TestCase("----", "----", null),
         new TestCase("---0", "---0", null),
         new TestCase("---1", "---1", null),
         new TestCase("--00", "--00", null),
         new TestCase("--01", "--01", null),
         new TestCase("1---", "1---", null),
         new TestCase("1--0", "1--0", null),
         new TestCase("1--1", "1--1", null),
         new TestCase("-1--", "-1--", null),
         new TestCase("-1-0", "-1-0", null),
         new TestCase("-1-1", "-1-1", null),
         new TestCase("-10-", "-10-", -10),
         new TestCase("-100", "-100", -100),
         new TestCase("-101", "-101", -101),
         new TestCase("-11-", "-11-", -11),
         new TestCase("-110", "-110", -110),
         new TestCase("-111", "-111", -111),
         new TestCase("11-0", "11-0", null),
         new TestCase("11-1", "11-1", null),
         new TestCase("10--", "10--", null),
         new TestCase("10-0", "10-0", null),
         new TestCase("10-1", "10-1", null),
         new TestCase("100-", "100-", 100),
         new TestCase("101-", "101-", 101),
         new TestCase("1-10", "1-10", null),
         new TestCase("1-11", "1-11", null),
         new TestCase("110-", "110-", 110),
         new TestCase("111-", "111-", 111),
         new TestCase("11--", "11--", null),
         new TestCase("1-0-", "1-0-", null),
         new TestCase("1-00", "1-00", null),
         new TestCase("1-01", "1-01", null),
         new TestCase("1-1-", "1-1-", null),
         new TestCase("01-0", "01-0", null),
         new TestCase("01-1", "01-1", null),
         new TestCase("010-", "010-", 10),
         new TestCase("011-", "011-", 11),
         new TestCase("00--", "00--", null),
         new TestCase("00-0", "00-0", null),
         new TestCase("00-1", "00-1", null),
         new TestCase("000-", "000-", null),
         new TestCase("0-10", "0-10", null),
         new TestCase("0-11", "0-11", null),
         new TestCase("-00-", "-00-", null),
         new TestCase("-000", "-000", null),
         new TestCase("-001", "-001", -1),
         new TestCase("001-", "001-", 1),
         new TestCase("0--0", "0--0", null),
         new TestCase("0--1", "0--1", null),
         new TestCase("0-00", "0-00", null),
         new TestCase("0-01", "0-01", null),
         new TestCase("-01-", "-01-", -1),
         new TestCase("-010", "-010", -10),
         new TestCase("-011", "-011", -11),
         new TestCase("--0_", "0--0", null),
         new TestCase("--1_", "0--1", null),
         new TestCase("---_", "0---", null),
         new TestCase("01-_", "001-", 1),
         new TestCase("1--_", "01--", null),
         new TestCase("1-0_", "01-0", null),
         new TestCase("1-1_", "01-1", null),
         new TestCase("10-_", "010-", 10),
         new TestCase("11-_", "011-", 11),
         new TestCase("-10_", "0-10", null),
         new TestCase("-11_", "0-11", null),
         new TestCase("00-_", "000-", null),
         new TestCase("-00_", "0-00", null),
         new TestCase("-01_", "0-01", null),
         new TestCase("0--_", "00--", null),
         new TestCase("0-0_", "00-0", null),
         new TestCase("0-1_", "00-1", null),
         new TestCase("-0-_", "0-0-", null),
         new TestCase("_0--", "-0--", null),
         new TestCase("_---", "----", null),
         new TestCase("_--0", "---0", null),
         new TestCase("_--1", "---1", null),
         new TestCase("_-0-", "--0-", null),
         new TestCase("_-00", "--00", null),
         new TestCase("_-01", "--01", null),
         new TestCase("_-1-", "--1-", null),
         new TestCase("_-10", "--10", null),
         new TestCase("_-11", "--11", null),
         new TestCase("_00-", "-00-", null),
         new TestCase("_01-", "-01-", -1),
         new TestCase("_10-", "-10-", -10),
         new TestCase("_11-", "-11-", -11),
         new TestCase("_0-0", "-0-0", null),
         new TestCase("_0-1", "-0-1", null),
         new TestCase("_1--", "-1--", null),
         new TestCase("_1-0", "-1-0", null),
         new TestCase("_1-1", "-1-1", null),
         new TestCase("1-__", "201-", 201),
         new TestCase("--__", "20--", null),
         new TestCase("-0__", "20-0", null),
         new TestCase("-1__", "20-1", null),
         new TestCase("0-__", "200-", 200),
         new TestCase("_--_", "0---", null),
         new TestCase("_-0_", "0--0", null),
         new TestCase("_-1_", "0--1", null),
         new TestCase("_0-_", "0-0-", null),
         new TestCase("__0-", "--0-", null),
         new TestCase("__1-", "--1-", null),
         new TestCase("__--", "----", null),
         new TestCase("__-0", "---0", null),
         new TestCase("__-1", "---1", null),
         new TestCase("_110", "0110", 110),
         new TestCase("001_", "0001", 1),
         new TestCase("0___", "2000", 2000),
         new TestCase("0__0", "2000", 2000),
         new TestCase("0__1", "2000", 2000),
         new TestCase("0_0_", "2000", 2000),
         new TestCase("0_00", "2000", 2000),
         new TestCase("0_01", "2000", 2000),
         new TestCase("0_1_", "2000", 2000),
         new TestCase("0_10", "2000", 2000),
         new TestCase("0_11", "2000", 2000),
         new TestCase("00_0", "2000", 2000),
         new TestCase("00_1", "2000", 2000),
         new TestCase("000_", "0000", null),
         new TestCase("01__", "2001", 2001),
         new TestCase("01_0", "2001", 2001),
         new TestCase("01_1", "2001", 2001),
         new TestCase("__0_", "2000", 2000),
         new TestCase("__00", "2000", 2000),
         new TestCase("__01", "2001", 2001),
         new TestCase("__1_", "2001", 2001),
         new TestCase("010_", "0010", 10),
         new TestCase("011_", "0011", 11),
         new TestCase("___0", "2000", 2000),
         new TestCase("___1", "2001", 2001),
         new TestCase("_0__", "2000", 2000),
         new TestCase("_0_0", "2000", 2000),
         new TestCase("_0_1", "2000", 2000),
         new TestCase("_00_", "2000", 2000),
         new TestCase("1_0_", "2001", 2001),
         new TestCase("1_00", "2001", 2001),
         new TestCase("1_01", "2001", 2001),
         new TestCase("1_1_", "2001", 2001),
         new TestCase("1_10", "2001", 2001),
         new TestCase("1_11", "2001", 2001),
         new TestCase("11_0", "2011", 2011),
         new TestCase("11_1", "2011", 2011),
         new TestCase("10_0", "2010", 2010),
         new TestCase("10_1", "2010", 2010),
         new TestCase("1__0", "2001", 2001),
         new TestCase("1__1", "2001", 2001),
         new TestCase("110_", "0110", 110),
         new TestCase("111_", "0111", 111),
         new TestCase("11__", "2011", 2011),
         new TestCase("100_", "0100", 100),
         new TestCase("10__", "2010", 2010),
         new TestCase("101_", "0101", 101),
         new TestCase("1___", "2001", 2001),
         new TestCase("00__", "2000", 2000),
         new TestCase("_10_", "2010", 2010),
         new TestCase("_100", "0100", 100),
         new TestCase("_101", "0101", 101),
         new TestCase("_11_", "2011", 2011),
         new TestCase("_111", "0111", 111),
         new TestCase("_01_", "2001", 2001),
         new TestCase("_000", "0000", null),
         new TestCase("_001", "0001", 1),
         new TestCase("_010", "0010", 10),
         new TestCase("_011", "0011", 11),
         new TestCase("_1_0", "2001", 2001),
         new TestCase("_1_1", "2001", 2001),
         new TestCase("_1__", "2001", 2001),
         new TestCase("__10", "2010", 2010),
         new TestCase("__11", "2011", 2011),
         new TestCase("0000", "0000", null),
         new TestCase("1110", "1110", 1110),
         new TestCase("1111", "1111", 1111),
         new TestCase("1100", "1100", 1100),
         new TestCase("1101", "1101", 1101),
         new TestCase("1010", "1010", 1010),
         new TestCase("1011", "1011", 1011),
         new TestCase("1000", "1000", 1000),
         new TestCase("1001", "1001", 1001),
         new TestCase("0100", "0100", 100),
         new TestCase("0101", "0101", 101),
         new TestCase("0010", "0010", 10),
         new TestCase("0011", "0011", 11),
         new TestCase("0001", "0001", 1),
         new TestCase("0110", "0110", 110),
         new TestCase("0111", "0111", 111),
         new TestCase("__", "08", 2008),
         new TestCase("_-", "--", 2000),
         new TestCase("-_", "0-", 2000),
         new TestCase("--", "--", 2000),
         new TestCase("-0", "-0", 2000),
         new TestCase("-1", "-1", -1),
         new TestCase("_0", "00", 2000),
         new TestCase("_1", "01", 2001),
         new TestCase("0-", "0-", 2000),
         new TestCase("1-", "1-", 2001),
         new TestCase("0_", "00", 2000),
         new TestCase("1_", "01", 2001),
         new TestCase("00", "00", 2000),
         new TestCase("01", "01", 2001),
         new TestCase("10", "10", 2010),
         new TestCase("11", "11", 2011),

         //previous century
         new TestCase("99__", "1999", 1999),
         new TestCase("_99_", "1999", 1999),
         new TestCase("-99_", "0-99", null),   
         new TestCase("99", "99", 1999)
      };

      boolean completedSuccessfully = true;
      ErrorManager.setHeadless();

      for(TestCase test : tests)
      {
         fmtFullYear = test.getOldBuf().length() == 4;

         char[][] newCompBufs = new char[3][];
         char[][] oldCompBufs = new char[3][];

         if(fmtFullYear)
         {
            newCompBufs[YEAR] = new char[4];
            oldCompBufs[YEAR] = new char[]{test.getOldBuf().charAt(0),
                                           test.getOldBuf().charAt(1),
                                           test.getOldBuf().charAt(2),
                                           test.getOldBuf().charAt(3)};
         }
         else
         {
            newCompBufs[YEAR] = new char[2];
            oldCompBufs[YEAR] = new char[]{test.getOldBuf().charAt(0),
                                           test.getOldBuf().charAt(1)};
         }

         // autocompletion test
         DateBuf buf = new DateBuf();
         buf.completeComponent(newCompBufs,oldCompBufs,YEAR);

         String newBuf = new String(newCompBufs[YEAR]);
         if(!test.getNewBuf().equals(newBuf))
         {
            System.out.println("Autocompletion error: '" + test.getOldBuf() +
                               "' is converted to '" + newBuf +
                               "' but it should be '" + test.getNewBuf() +
                               "'");
            completedSuccessfully = false;
         }

         // conversion test
         Integer year = null;
         buf.compBufs[DAY]   = new char[] {'0','1'};
         buf.compBufs[MONTH] = new char[] {'0','1'};
         buf.compBufs[YEAR]  = newCompBufs[YEAR];
         try
         {
            year = buf.toDate().getYear();
         }
         catch (DisplayFormatCheckException ex)
         {
            // year = null
         }
         catch (ErrorConditionException ex)
         {
            // year = null
         }

         if((year == null && test.getYear() != null)      ||
            (year != null && !year.equals(test.getYear())))
         {
            System.out.println("Date conversion error: '" + newBuf +
                               "' is converted to '" + year +
                               "' but it should be '" + test.getYear() +
                               "'");
            completedSuccessfully = false;
         }
      }

      if(completedSuccessfully)
         System.out.println("All tests completed successfully");
   }

   /**
    * Start all year autocompletion and its string-to-number conversion
    * testcases.
    * 
    * @param   args
    *          Command line arguments.
    * @throws  DisplayFormatParsingException
    */
   public static void main(String[] args)
   throws DisplayFormatParsingException
   {
      DateFormat f = new DateFormat("99/99/9999");
      f.test();
   }
}
