/*
** Module   : SessionExports.java
** Abstract : interface which describes all session-attribute API's exported
**            by client for use by server  
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --------------------------------Description---------------------------------
** 001 OM  20130510 Initial commit. Only date/time-related attributes exposed.
** 002 OM  20130607 Added windowing year attribute (YEAR-OFFSET).
*/

package com.goldencode.p2j.util;

/**
 * Describes all SESSION-related API's exported by client for use by server.
 * <p>
 * Using this interface, all SESSION attributes are available for query on client side.
 * 
 * @author om
 */
public interface SessionExports
{
   /** 
    * The SESSION:DATE-FORMAT sets the order of date components for conversion to/from strings.
    * <p>
    * The order of the 3 date sub-components as a 3 character string using "M", "D" and "Y" once 
    * each.  The order from index 0 to index 2 represents the left to right ordering of the 
    * components. So the leftmost date component will be defined by the character at index 
    * position 0.
    * <p>
    * The value set is specific to the current user's context.
    *
    * @param    dateFormat
    *           Defines the order of date components for conversion to/from strings.
    */
   public void setDateOrder(character dateFormat);
   
   /**
    * Sets the SESSION:TIMEZONE used as the default offset in minutes from
    * UTC for this user's context.
    *
    * @param    offset
    *           The new timezone offset which must be between -840 and 840 inclusive. 
    *           The value may also be <code>unknown</code> which is the equivalent of resetting
    *           this attribute to the default.
    */
   public void setSessionTimeZone(NumberType offset);
   
   /**
    * Sets the <code>TIME-SOURCE</code> attribute for the user's context.
    * This attribute is can be read and written but it is not honored for date/time processing in
    * P2J.  The intention of the attribute is to define the system from which time values will be
    * obtained. In P2J, time values are always obtained from the application server.
    *
    * @param    timeSource
    *           The new value to set. If <code>null</code>, then the empty string will be used,
    *           which in the 4GL has the same affect as "local".
    */
   public void setTimeSource(character timeSource);

   /**
    * Sets the <code>DISPLAY-TIMEZONE</code> attribute for the user's context.
    * This attribute is used for displaying DATETIME-TZ values that do not have 
    * the timezone specified in the display format. 
    *
    * @param    displayTimezone
    *           The new value to set.
    */
   public void setDisplayTimeZone(NumberType displayTimezone);

   /**
    * Implementation of SESSION:YEAR-OFFSET attribute setter. Sets the 
    * configured start year for the two-digit Y2K windowing range. This is
    * the equivalent of the Progress -yy command line parameter. The 
    * default value is 1950 unless otherwise specified in the directory (see
    * {@link date.getWindowingYear}).
    * 
    * @param   year
    *          The new start year for the two-digit windowing range.
    */
   public void setYearOffset(NumberType year);
}
